#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from StanfordDependencies import StanfordDependencies

def print_sentences(sentences):
    for sentence in sentences:
        print '---'
        for token in sentence:
            print token
    print

if __name__ == "__main__":
    # simple testing script -- ideally will be replaced with proper
    # unit tests

    for backend in ('jpype', 'subprocess'):
        print 'Backend:', backend
        sd = StanfordDependencies.get_instance(backend=backend)
        sents = sd.convert_trees(['(S1 (NP (DT a) (NN cow)))',
                                  '(S1 (NP (DT some) (NNS cows)))',
                                  '(S1 (NP (DT some) (JJ blue) (NN moose)))'],
                                 representation='basic')
        print_sentences(sents)

        tree = '(S1 (NP (NP (NP (DT A) (NN cat)) (CC and) ' + \
               '(NP (DT a) (NN mouse))) (. .)))'
        for representation in ('basic', 'collapsed', 'CCprocessed',
                               'collapsedTree'):
            print '%s representation' % representation
            print_sentences(sd.convert_trees([tree],
                                             representation=representation))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import tempfile
from .StanfordDependencies import StanfordDependencies
from .Token import Token

JAVA_CLASS_NAME = 'edu.stanford.nlp.trees.EnglishGrammaticalStructure'

def read_conll_sentences(stream, deps_to_exclude=None):
    """stream is an iterable over strings. Returns a list of sentences
    where a sentence is a list of Token objects. deps_to_exclude
    is a collection of dependency labels to skip."""
    deps_to_exclude = deps_to_exclude or []
    current_sentence = []
    sentences = []
    def flush():
        if current_sentence:
            sentences.append(list(current_sentence))
            del current_sentence[:]
    for line in stream:
        line = line.strip()
        if line:
            token = Token.from_string(line)
            if token.deprel in deps_to_exclude:
                continue
            current_sentence.append(token)
        else:
            flush()
    flush()
    return sentences

class SubprocessBackend(StanfordDependencies):
    """Interface to Stanford Dependencies via subprocesses. This means
    that each call opens a pipe to Java. It has the advantage that
    it should work out of the box if you have Java but is slower than
    other backends. As such, convert_trees() will be more efficient than
    convert_tree() for this backend."""
    def __init__(self, jar_filename=None, download_if_missing=False,
                 version=None, java_command='java'):
        """java_command is the path to a java binary."""
        StanfordDependencies.__init__(self, jar_filename, download_if_missing,
                                      version)
        self.java_command = java_command
    def convert_trees(self, ptb_trees, java='java', representation='basic',
                      include_punct=True, include_erased=False):
        """Convert a list of Penn Treebank formatted trees (ptb_trees)
        into Stanford Dependencies. The dependencies are represented
        as a list of sentences, where each sentence is itself a list of
        Token objects.

        Currently supported representations are 'basic', 'collapsed',
        'CCprocessed', and 'collapsedTree' which behave the same as they
        in the CoreNLP command line tools."""
        with tempfile.NamedTemporaryFile() as input_file:
            for ptb_tree in ptb_trees:
                input_file.write(ptb_tree + '\n')
            input_file.flush()

            command = [self.java_command,
                       '-ea',
                       '-cp', self.jar_filename,
                       JAVA_CLASS_NAME,
                       '-' + representation,
                       '-treeFile', input_file.name,
                       '-conllx']
            sd_process = subprocess.Popen(command, stdout=subprocess.PIPE,
                                          stderr=subprocess.PIPE)
            return_code = sd_process.wait()
            stderr = sd_process.stderr.read()
            stdout = sd_process.stdout.read()
            if return_code:
                if 'Unsupported major.minor version' in stderr:
                    self.java_is_too_old()
                raise RuntimeError("Bad exit code from Stanford CoreNLP")

        deps_to_exclude = []
        if not include_erased:
            deps_to_exclude.append('erased')
        if not include_punct:
            deps_to_exclude.append('punct')
        sentences = read_conll_sentences(stdout.splitlines(), deps_to_exclude)

        if len(sentences) != len(ptb_trees):
            raise RuntimeError("Only got %d sentences from Stanford Dependencies when given %d trees." % (len(sentences), len(ptb_trees)))
        return sentences
    def convert_tree(self, ptb_tree, **kwargs):
        """Converts a single Penn Treebank format tree to Stanford
        Dependencies. See convert_trees for more details."""
        return self.convert_trees([ptb_tree], **kwargs)[0]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jpype
from .StanfordDependencies import StanfordDependencies
from .Token import Token

class JPypeBackend(StanfordDependencies):
    """Faster backend than SubprocessBackend but requires you to install
    jpype ('pip install JPype1', not 'JPype'). May be less stable. There's
    no speed benefit of using convert_trees() over convert_tree() for this
    backend. In terms of output, should be identical to SubprocessBackend
    except that all string fields will be unicode. Additionally, has
    the option to run the lemmatizer (see convert_tree())."""
    def __init__(self, jar_filename=None, download_if_missing=False,
                 version=None, extra_jvm_args=None, start_jpype=True):
        StanfordDependencies.__init__(self, jar_filename, download_if_missing,
                                      version)
        if start_jpype:
            jpype.startJVM(jpype.getDefaultJVMPath(), '-ea',
                           '-Djava.class.path=' + self.jar_filename,
                           *(extra_jvm_args or []))
        self.corenlp = jpype.JPackage('edu').stanford.nlp
        try:
            self.acceptFilter = self.corenlp.util.Filters.acceptFilter()
        except TypeError:
            # this appears to be caused by a mismatch between CoreNLP and JRE
            # versions since this method changed to return a Predicate.
            version = jpype.java.lang.System.getProperty("java.version")
            print "Your Java version:", version
            if version.startswith('1.7'):
                print "The last CoreNLP release for Java 1.7 was 3.4.1"
                print "Try using: StanfordDependencies.get_instance(backend='jpype', version='3.4.1')"
            print
            self.java_is_too_old()
        trees = self.corenlp.trees
        self.treeReader = trees.Trees.readTree
        self.grammaticalStructure = trees.EnglishGrammaticalStructure
        self.stemmer = self.corenlp.process.Morphology.stemStaticSynchronized
    def convert_tree(self, ptb_tree, representation='basic',
                     include_punct=True, include_erased=False,
                     add_lemmas=False):
        """Arguments are as in StanfordDependencies.convert_trees but with
        the addition of add_lemmas. If add_lemmas=True, we will run the
        Stanford CoreNLP lemmatizer and fill in the lemma field."""
        tree = self.treeReader(ptb_tree)
        if include_punct:
            egs = self.grammaticalStructure(tree, self.acceptFilter)
        else:
            egs = self.grammaticalStructure(tree)

        if representation == 'basic':
            deps = egs.typedDependencies()
        elif representation == 'collapsed':
            deps = egs.typedDependenciesCollapsed(True)
        elif representation == 'CCprocessed':
            deps = egs.typedDependenciesCCprocessed(True)
        elif representation == 'collapsedTree':
            deps = egs.typedDependenciesCollapsedTree()
        else:
            raise ValueError("Unknown representation: %r" % representation)

        head_and_deprel = {}
        for dep in deps:
            head_and_deprel[dep.dep().index()] = (dep.gov().index(),
                                                  dep.reln().toString())

        words = tree.taggedYield()
        tokens = []
        for index, word in enumerate(words, 1):
            if index in head_and_deprel:
                head, deprel = head_and_deprel[index]
            elif include_erased:
                head, deprel = 0, 'erased'
            else:
                continue
            form = word.value()
            tag = word.tag()
            if add_lemmas:
                lemma = self.stemmer(form, tag).word()
            else:
                lemma = None
            token = Token(index=index, form=form, lemma=lemma, cpos=tag,
                          pos=tag, feats=None, head=head, deprel=deprel,
                          phead=None, pdeprel=None)
            tokens.append(token)
        return tokens



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

# ideally, this will be set to the latest version of CoreNLP
DEFAULT_CORENLP_VERSION = '3.5.0'

# where we store downloaded jar files
INSTALL_DIR = '~/.local/share/pystanforddeps'

class StanfordDependencies:
    """Abstract base class for converting Penn Treebank trees to Stanford
    Dependencies. To actually use this, you'll want to instantiate one
    of the backends. The easiest way to do this is via the get_instance()
    class method.

    If you do not currently have the appropriate Java jar files, the
    download_if_missing flag in the constructor will help you fetch them.
    In this case, you can set version to a string with the CoreNLP version
    you'd like. If unset, it will default to DEFAULT_CORENLP_VERSION.

    Subclasses should (at minimum) override the convert_tree method. They
    may also want to override convert_trees if they require batch
    operation. They may also add their own custom keyword arguments to
    __init__(), convert_tree(), and convert_trees()."""
    __metaclass__ = ABCMeta
    def __init__(self, jar_filename=None, download_if_missing=False,
                 version=None):
        """jar_filename should be the path to a Java jar file with
        classfiles from Stanford CoreNLP or Stanford Parser.

        If download_if_missing is True, it will automatically download
        a jar file and store it locally. By default it will use
        DEFAULT_CORENLP_VERSION but will use the version flag if
        that argument is specified."""
        if not (jar_filename or version or download_if_missing):
            raise ValueError("Must set either jar_filename, version, or download_if_missing to True.")

        self.jar_filename = jar_filename
        if not self.jar_filename:
            version = version or DEFAULT_CORENLP_VERSION
            filename = 'stanford-corenlp-%s.jar' % version
            self.jar_filename = self.setup_and_get_default_path(filename)
            if download_if_missing:
                self.download_if_missing(version)
    def convert_trees(self, ptb_trees, representation='basic', **kwargs):
        """Convert a list of Penn Treebank formatted trees (ptb_trees)
        into Stanford Dependencies. The dependencies are represented
        as a list of sentences, where each sentence is itself a list of
        Token objects.

        Currently supported representations are 'basic', 'collapsed',
        'CCprocessed', and 'collapsedTree' which behave the same as they
        in the CoreNLP command line tools.

        Additionally, most backends accept at least two additional
        arguments: include_punct (if false, punctuation tokens
        will not be included) and include_erased (if false and
        your representation might erase tokens, those tokens will be
        skipped). See documentation on your backend to see if it supports
        more options."""
        return [self.convert_tree(ptb_tree, representation=representation,
                                  **kwargs)
                for ptb_tree in ptb_trees]

    @abstractmethod
    def convert_tree(self, ptb_tree, representation='basic', **kwargs):
        """Converts a single Penn Treebank format tree to Stanford
        Dependencies. See convert_trees for more details."""

    def setup_and_get_default_path(self, jar_base_filename):
        """Determine the user-specific install path for the Stanford
        Dependencies jar if the jar_url is not specified and ensure that
        it is writable (that is, make sure the directory exists). Returns
        the full path for where the jar file should be installed."""
        import os
        import os.path

        install_dir = os.path.expanduser(INSTALL_DIR)
        try:
            os.makedirs(install_dir)
        except OSError:
            pass
        jar_filename = os.path.join(install_dir, jar_base_filename)
        return jar_filename
    def download_if_missing(self, version=None, verbose=True):
        """Download the jar for version into the jar_filename specified
        in the constructor. Will not overwrite jar_filename if it already
        exists. version defaults to DEFAULT_CORENLP_VERSION (ideally the
        latest but we can't guarantee that since PyStanfordDependencies
        is distributed separately)."""
        import os.path
        if not os.path.exists(self.jar_filename):
            jar_url = self.get_jar_url(version)
            if verbose:
                print "Downloading %r -> %r" % (jar_url, self.jar_filename)
            import urllib
            urllib.urlretrieve(jar_url, filename=self.jar_filename)

    @staticmethod
    def java_is_too_old():
        """Warn the user that their JRE is too old to handle the requested
        version of Stanford CoreNLP."""
        raise RuntimeError("Your Java runtime is too old (must be 1.8+ to use CoreNLP version 3.5.0 or later)")

    @staticmethod
    def get_jar_url(version=None):
        """Get the URL to a Stanford CoreNLP jar file with a specific
        version. These jars come from Maven since the maven version is
        smaller than the full CoreNLP distributions. Defaults to
        DEFAULT_CORENLP_VERSION."""
        version = version or DEFAULT_CORENLP_VERSION
        jar_filename = 'stanford-corenlp-%s.jar' % version
        return 'http://search.maven.org/remotecontent?filepath=' + \
            'edu/stanford/nlp/stanford-corenlp/%s/%s' % (version,
                                                         jar_filename)

    @staticmethod
    def get_instance(jar_filename=None, version=None,
                     download_if_missing=True, backend='jpype',
                     **extra_args):
        extra_args.update(jar_filename=jar_filename,
                          download_if_missing=download_if_missing,
                          version=version)
        if backend == 'jpype':
            try:
                from JPypeBackend import JPypeBackend
                return JPypeBackend(**extra_args)
            except ImportError:
                import warnings
                warnings.warn('Error importing JPypeBackend, ' +
                              'falling back to SubprocessBackend.')

                # fall back to subprocess backend which should work
                # more generally
                backend = 'subprocess'

        if backend == 'subprocess':
            from SubprocessBackend import SubprocessBackend
            return SubprocessBackend(**extra_args)

        raise ValueError("Unknown backend: %r (known backends: 'subprocess' and 'jpype')" % backend)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

# CoNLL-X field names
FIELD_NAMES = ('index', 'form', 'lemma', 'cpos', 'pos', 'feats', 'head',
               'deprel', 'phead', 'pdeprel')

class Token(namedtuple('Token', FIELD_NAMES)):
    """A CoNLL-X style dependency token. Fields include:
    - form (the word form)
    - lemma (the word's base form or lemma) -- empty for SubprocessBackend
    - pos (part of speech tag)
    - index (index of the token in the sentence)
    - head (index of the head of this token), and
    - deprel (the dependency relation between this token and its head)

    There are other fields but they typically won't be populated by
    StanfordDependencies.

    See http://ilk.uvt.nl/conll/#dataformat for a complete description."""
    def __repr__(self):
        # slightly different from the official tuple __repr__ in that
        # we skip any fields with None as their value
        items = [(field, getattr(self, field, None)) for field in FIELD_NAMES]
        fields = ['%s=%r' % (k, v) for k, v in items if v is not None]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(fields))
    @classmethod
    def from_string(this_class, text):
        fields = text.split('\t')
        fields[0] = int(fields[0]) # index
        fields[6] = int(fields[6]) # head index
        fields = [value if value != '_' else None for value in fields]
        return this_class(**dict(zip(FIELD_NAMES, fields)))



#!/usr/bin/env python

from distutils.core import setup

setup(name='PyStanfordDependencies',
      version='0.1',
      description='Python interface for converting Penn Treebank trees to Stanford Dependencies',
      long_description='See http://github.com/dmcc/PyStanfordDependencies',
      author='David McClosky',
      author_email='notsoweird+pystanforddependencies@gmail.com',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: Apache Software License',
          'Natural Language :: English',
          'Operating System :: POSIX',
          'Topic :: Scientific/Engineering :: Artificial Intelligence',
      ],
      url='http://github.com/dmcc/PyStanfordDependencies',
      license='Apache 2.0',
      platforms=['POSIX'],
      packages=['StanfordDependencies'])
